--******************************************************************
--  CPU.vhd
--******************************************************************
library IEEE;
use IEEE.STD_LOGIC_1164.ALL;
use IEEE.STD_LOGIC_SIGNED.ALL;

use work.cpupackage.ALL;
-- synopsys translate_off
use work.debug.ALL;
-- synopsys translate_on

entity cpu32 is
    Port ( reset : in    std_logic;
           clock : in    std_logic;
           pcOut : out   std_logic_vector(31 downto 0)
         );
end entity cpu32;
--
--************************************************************************
--
architecture Behavioral of cpu32 is

-- ALU signals
signal  aluOp               : std_logic_vector( 2 downto 0);
signal  aluOperand1         : std_logic_vector(31 downto 0);
signal  aluOperand2         : std_logic_vector(31 downto 0);
signal  aluDataOut          : std_logic_vector(31 downto 0);
signal  Z,N,V,C             : std_logic;

-- Instruction register
signal ir                   : std_logic_vector(31 downto 0);
signal loadIR               : std_logic;

-- Values calculated from IR
signal branchOffset           : std_logic_vector(31 downto 0);
signal signedImmediateValue   : std_logic_vector(31 downto 0);
signal unsignedImmediateValue : std_logic_vector(31 downto 0);

-- Registers and intermediate values
signal pc                   : std_logic_vector(31 downto 0);
signal loadPC               : std_logic;
signal PCSource             : PCSourceT;

-- Port A of Register file (write)
signal regAAddr             : std_logic_vector( 4 downto 0);
signal regAWrite            : std_logic;
signal regADataIn           : std_logic_vector(31 downto 0);
signal RegASource           : RegASourceT; 

-- Port B of Register file (read)
signal regBAddr             : std_logic_vector( 4 downto 0);
signal regBDataOut          : std_logic_vector(31 downto 0);

-- Port C of Register file (read)
signal regCAddr             : std_logic_vector( 4 downto 0);
signal regCDataOut          : std_logic_vector(31 downto 0);

-- Data Memory
signal dataMemAddr          : std_logic_vector(11 downto 0);
signal dataMemWrite         : std_logic;
signal dataMemDataIn        : std_logic_vector(31 downto 0);
signal dataMemDataOut       : std_logic_vector(31 downto 0);

-- Code Memory
signal codeMemAddr          : std_logic_vector(11 downto 0);
signal codeMemDataOut       : std_logic_vector(31 downto 0);
signal codeMemRead          : std_logic;

--**********************************************************
--* Modification for Bug 1
--**********************************************************
-- Condition Code Register
signal CCR                  : std_logic_vector(3 downto 0);
signal saveCCR              : std_logic;
-- End modification for Bug 1
--**********************************************************

--**********************************************************
--* Modification for Bug 2
--**********************************************************
-- Signal to store PC when BSR & JMP
signal savePC               : std_logic;
-- End modification for Bug 2
--**********************************************************

constant byteAddressable : boolean := false;

-- 1024 words of Data memory
component datamemory
   port (
      clk  : in  std_logic;
      we   : in  std_logic;
      addr : in  std_logic_vector(9 downto 0);
      din  : in  std_logic_vector(31 downto 0);
      dout : out std_logic_vector(31 downto 0));
end component datamemory;

-- 1024 words of Code memory
component codememory
   port (
      clk  : in  std_logic;
      en   : in  std_logic;
      addr : in  std_logic_vector(9 downto 0);
      dout : out std_logic_vector(31 downto 0));
end component codememory;

--************************************************************************
begin

   pcOut <= pc; -- to stop design being optimized away!

   --==============================================
   -- PC source data path
   --
   pcProcess:
   process (reset, clock, ir)
   begin

      -- 32-bit signed-extended branch offset from ir
      branchOffset <= sex(branchOffset'length, ir_branchOffset(ir));

      if (reset = '1') then
         pc  <= (others => '0');
      elsif (clock'event and (clock = '1')) then
         if (loadPC = '1') then
            case pcSource is
               when branchPC =>  pc <= PC + branchOffset;
               when nextPC   =>  pc <= PC + 4;
               when jumpPC   =>  pc <= aluDataOut;
            end case;
         end if;
       end if;
   end process;

   --==============================================
   -- Main data path
   --
   dataMemAddr    <= aluDataOut(dataMemAddr'left downto 0);
   dataMemDataIn  <= regCDataOut;

   codeMemAddr    <= PC(codeMemAddr'length-1 downto 0);
   ir             <= codeMemDataOut;
   
   codeMemRead    <= loadIR;

   --=============================================
   -- Register File control
   --
   regControl:
   process (ir, RegASource, aluDataOut, dataMemDataOut, PC)
   begin
      regAAddr   <= ir_regA(ir);
      regBAddr   <= ir_regB(ir);
      regCAddr   <= ir_regC(ir);
      
      --**********************************************************
      --* Modification for Bug 2
      --**********************************************************
      -- Save the PC counter if BSR, JMP occur
      if (savePC = '1') then
        regADataIn <= PC;
        regAAddr <= "11111";
      else
      -- End modification for Bug 2
      --**********************************************************
        case RegASource is
          when aluOut      => regADataIn <= aluDataOut;      
          when dataMemOut  => regADataIn <= dataMemDataOut;
        end case;
      end if;
   end process regControl;

   --=============================================
   -- ALU operand source multiplexing
   -- ALU operation control
   --
   aluControl:
   process ( ir, signedImmediateValue, regBDataOut, regCDataOut )

   begin

      if (ir_op(ir) = "000") then
         aluOp  <= ir_aluOp(ir);
      else
         aluOp  <= ALUopAdd; -- add for all other opcodes
      end if;

      -- Set up operand 1 to ALU
      aluOperand1 <= regBDataOut;

      -- Set up operand 2 to ALU
      signedImmediateValue   <= sex(signedImmediateValue'length, ir_immediate16(ir));
      unsignedImmediateValue <= zex(signedImmediateValue'length, ir_immediate16(ir));
      
      if (ir_op(ir) = "000") then
         aluOperand2  <= regCDataOut;
      elsif (ir_op(ir) = "001") and (ir_aluOp(ir) /= ALUopEor) then
         -- 32-bit zero-extended immediate value from ir
         aluOperand2  <= signedImmediateValue;
      else  
         -- 32-bit sign-extended immediate value from ir
         aluOperand2  <= unsignedImmediateValue;
      end if;

   end process aluControl;
   
   --*************************************************************
   --* Modification for Bug 1
   --*************************************************************
   -- Condition Code Register implementation
   process(clock,reset)
   begin 
     if reset = '1' then
       CCR <= (others => '0');
     elsif clock'event and clock = '1' then
       if saveCCR = '1' then
         CCR <= Z & N & V & C;
       end if;
     end if;
   end process;
   -- End of modification for Bug 1
   --*************************************************************
    
   --=============================================================
   -- ALU instantiation
   --
   theALU:
   entity work.alu(Behavioral)
      Port Map (
         aluOp     => aluOp,
         operand1  => aluOperand1,
         operand2  => aluOperand2,
         aluOutput => aluDataOut,
         Z         => Z,
         N         => N,
         V         => V,
         C         => C
         );

   --=============================================================
   -- Register Bank instantiation
   --
   registerBank :         
   entity work.Registers(DistributedRAM) -- use for synthesis
--   entity work.Registers(Muxes)        -- simulation only
      Generic Map (
         dataWidth    => 32,
         addressWidth => 5
         )
      Port Map ( 
         reset     => reset,
         clock     => clock,
         -- Port A (write)
         writeA    => regAWrite,
         addressA  => regAAddr,
         dataInA   => regADataIn,
         -- Port B (read)
         addressB  => regBAddr,
         dataOutB  => regBDataOut,
         -- Port C (read)
         addressC  => regCAddr,
         dataOutC  => regCDataOut
         );

   --=============================================================
   -- Data Memory instantiation
   -- 1024 words (A1..0 not used)
   dataMem:
   datamemory port map (
      clk  => clock,
      addr => dataMemAddr(11 downto 2),
      din  => dataMemDataIn,
      dout => dataMemDataOut,
      we   => dataMemWrite
      );

   --=============================================================
   -- Code Memory instantiation
   -- 1024 words (A1..0 not used)
   codeMem:
   codememory port map (
      clk  => clock,
      addr => codeMemAddr(11 downto 2),
      dout => codeMemDataOut,
      en   => codeMemRead
      );

   --=============================================================
   -- Controller state machine
   --
   theController:
   entity work.Control
      Port map (
         reset        => reset,
         clock        => clock,
         ir           => ir,
         Z            => CCR(3),
         N            => CCR(2),
         V            => CCR(1),
         C            => CCR(0),
         regAWrite    => regAWrite,
         RegASource   => RegASource, 
         loadPC       => loadPC,
         loadIR       => loadIR,
         dataMemWrite => dataMemWrite,
         PCSource     => PCSource,
         --*************************************
         --* Modification for Bug 1 & 2
         --*************************************
         saveCCR      => saveCCR,
         savePC       => savePC
         -- End of modification for Bug 1 & 2
         --*************************************
         );

end architecture Behavioral;